#include <iostream.h>
#include <fstream.h>
#include <iomanip.h>

#define MY_EOF_CHAR	256

bool putback_char_p = 0;
int putback_char;

int prev_char_num;
int prev_line_num;
int char_num = 1;
int line_num = 1;
int s_char_num;
int s_line_num;

int next_char (ifstream &fstream) {
  int ch;
  if (putback_char_p) {
    putback_char_p = 0;
    ch = putback_char;
  }
  else
    ch = fstream.get();

  if (ch == EOF) ch = MY_EOF_CHAR;

  prev_char_num = char_num;
  prev_line_num = line_num;

  if (ch == '\n') {
    line_num++;
    char_num = 1;
  }
  else
    char_num++;

  return ch;
}

bool put_back_char (int ch) {
  if (putback_char_p) {
    cerr << "Error!  Cannot put back twice!" << endl;
    return 0;
  }
  else {
    char_num = prev_char_num;
    line_num = prev_line_num;
    putback_char_p = 1;
    putback_char = ch;
    return 1;
  }
}

#define T_NONE		0
#define T_EOF           1
#define T_PLUS          11
#define T_TIMES         12
#define T_EQUALS        13
#define T_ASSIGN        21
#define T_IDENT         31

int yytext_loc = 0;
char yytext[1000];

ifstream yyin;

#define letter_p(ch)	((((ch) >= 'A') && ((ch) <= 'Z')) || (((ch) >= 'a') && ((ch) <= 'z')))
#define digit_p(ch)	(((ch) >= '0') && ((ch) <= '9'))

int yylex () {
  int ch;

  yytext_loc = 0;
  yytext[yytext_loc] = '\0';

  do {
    s_line_num = line_num;
    s_char_num = char_num;
    ch = next_char(yyin);
    if (letter_p(ch)) {
      yytext[0] = (char) ch;
      yytext[1] = '\0';
      yytext_loc = 1;
      do {
        ch = next_char(yyin);
        if (letter_p(ch) || digit_p(ch)) {
          yytext[yytext_loc] = (char) ch;
          yytext_loc++;
          yytext[yytext_loc] = '\0';
        }
        else {
          put_back_char(ch);
          return T_IDENT;
        }
      } while (1);
    }
    else if (ch == MY_EOF_CHAR) {
      strcpy(yytext,"");
      return T_EOF;
    }
    else {
      switch ((char) ch) {
        case '+': strcpy(yytext,"+"); return T_PLUS;
        case '*': strcpy(yytext,"*"); return T_TIMES;
        case ' ': case '\t': case '\n': break;
        case '=':
          ch = next_char(yyin);
          if (ch == '=') {
            strcpy(yytext,"==");
            return T_EQUALS;
          }
          else {
            put_back_char(ch);
            strcpy(yytext,"=");
            return T_ASSIGN;
          }
          break;
        default: cout << "Error!  Unexpected char (" << (char) ch << ") at line " << prev_line_num << ", char " << prev_char_num << "!" << endl;
      }
    }
  } while (1);
}

int main (int argc, char *argv[]) {
  if (argc != 2) {
    cout << "Useage: assign_scan <file>" << endl;
    return -1;
  }

  yyin.open(argv[1]);

  if (!yyin) {
    cout << "Error!  Unable to open file " << argv[1] << endl;
    return -1;
  }

  int token_num;
  do {
    token_num = yylex();

    cout << "Token: ";
    switch (token_num) {
      case T_EOF: cout << "EOF"; break;
      case T_PLUS: cout << "+"; break;
      case T_TIMES: cout << "*"; break;
      case T_EQUALS: cout << "=="; break;
      case T_ASSIGN: cout << "="; break;
      case T_IDENT: cout << "Ident(" << yytext << ")"; break;
    }
    cout << " found starting at line " << s_line_num << " char " << s_char_num << "\n";
  } while (token_num != T_EOF);

  return 0;
}
